@@ -1,6 +1,6 @@
 # encoding:utf-8
 #--
-# Copyright (C) 2006-2011 Bob Aman
+# Copyright (C) 2006-2013 Bob Aman
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -162,6 +162,15 @@ module Addressable
       return nil unless uri
       # If a URI object is passed, just return itself.
       return uri.dup if uri.kind_of?(self)
+
+      # If a URI object of the Ruby standard library variety is passed,
+      # convert it to a string, then parse the string.
+      # We do the check this way because we don't want to accidentally
+      # cause a missing constant exception to be thrown.
+      if uri.class.name =~ /^URI\b/
+        uri = uri.to_s
+      end
+
       if !uri.respond_to?(:to_str)
         raise TypeError, "Can't convert #{uri.class} into String."
       end
@@ -415,8 +424,13 @@ module Addressable
           "Expected Class (String or Addressable::URI), " +
           "got #{return_type.inspect}"
       end
-      result = uri.gsub(/%[0-9a-f]{2}/i) do |sequence|
+      uri = uri.dup
+      # Seriously, only use UTF-8. I'm really not kidding!
+      uri.force_encoding("utf-8") if uri.respond_to?(:force_encoding)
+      leave_encoded.force_encoding("utf-8") if leave_encoded.respond_to?(:force_encoding)
+      result = uri.gsub(/%[0-9a-f]{2}/iu) do |sequence|
         c = sequence[1..3].to_i(16).chr
+        c.force_encoding("utf-8") if c.respond_to?(:force_encoding)
         leave_encoded.include?(c) ? sequence : c
       end
       result.force_encoding("utf-8") if result.respond_to?(:force_encoding)
@@ -496,7 +510,7 @@ module Addressable
       end
       if character_class.kind_of?(String)
         leave_re = if leave_encoded.length > 0
-          character_class << '%'
+          character_class = "#{character_class}%" unless character_class.include?('%')
 
           "|%(?!#{leave_encoded.chars.map do |char|
             seq = char.unpack('C*').map { |c| '%02x' % c }.join
@@ -522,6 +536,9 @@ module Addressable
       rescue ArgumentError
         encoded = self.encode_component(unencoded)
       end
+      if encoded.respond_to?(:force_encoding)
+        encoded.force_encoding(Encoding::UTF_8)
+      end
       return encoded
     end
 
@@ -1049,8 +1066,8 @@ module Addressable
           result = ::Addressable::IDNA.to_ascii(
             URI.unencode_component(self.host.strip.downcase)
           )
-          if result[-1..-1] == "."
-            # Trailing dots are unnecessary
+          if result =~ /[^\.]\.$/
+            # Single trailing dots are unnecessary.
             result = result[0...-1]
           end
           result
@@ -1089,12 +1106,32 @@ module Addressable
     end
 
     ##
+    # This method is same as URI::Generic#host except
+    # brackets for IPv6 (and 'IPvFuture') addresses are removed.
+    #
     # @see Addressable::URI#host
-    alias_method :hostname, :host
+    #
+    # @return [String] The hostname for this URI.
+    def hostname
+      v = self.host
+      /\A\[(.*)\]\z/ =~ v ? $1 : v
+    end
 
     ##
+    # This method is same as URI::Generic#host= except
+    # the argument can be a bare IPv6 address (or 'IPvFuture').
+    #
     # @see Addressable::URI#host=
-    alias_method :hostname=, :host=
+    #
+    # @param [String, #to_str] new_hostname The new hostname for this URI.
+    def hostname=(new_hostname)
+      if new_hostname && !new_hostname.respond_to?(:to_str)
+        raise TypeError, "Can't convert #{new_hostname.class} into String."
+      end
+      v = new_hostname ? new_hostname.to_str : nil
+      v = "[#{v}]" if /\A\[.*\]\z/ !~ v && /:/ =~ v
+      self.host = v
+    end
 
     ##
     # The authority component for this URI.
@@ -1175,7 +1212,7 @@ module Addressable
 
     ##
     # The origin for this URI, serialized to ASCII, as per
-    # draft-ietf-websec-origin-00, section 5.2.
+    # RFC 6454, section 6.2.
     #
     # @return [String] The serialized origin.
     def origin
@@ -1429,20 +1466,16 @@ module Addressable
     # The query component for this URI, normalized.
     #
     # @return [String] The query component, normalized.
-    def normalized_query
-      self.query && @normalized_query ||= (begin
-        modified_query_class = Addressable::URI::CharacterClasses::QUERY
-        # Make sure possible key-value pair delimiters are escaped.
-        modified_query_class = modified_query_class.sub("\\&", "")
-        modified_query_class = modified_query_class.sub("\\;", "")
-        (self.query.split("&", -1).map do |pair|
-          Addressable::URI.normalize_component(
-            pair,
-            modified_query_class,
-            '+'
-          )
-        end).join("&")
-      end)
+    def normalized_query(*flags)
+      modified_query_class = Addressable::URI::CharacterClasses::QUERY.dup
+      # Make sure possible key-value pair delimiters are escaped.
+      modified_query_class.sub!("\\&", "").sub!("\\;", "")
+      pairs = (self.query || "").split("&", -1)
+      pairs.sort! if flags.include?(:sorted)
+      component = (pairs.map do |pair|
+        Addressable::URI.normalize_component(pair, modified_query_class, "+")
+      end).join("&")
+      component == "" ? nil : component
     end
 
     ##
@@ -1623,10 +1656,11 @@ module Addressable
     # @return [String] The fragment component, normalized.
     def normalized_fragment
       self.fragment && @normalized_fragment ||= (begin
-        Addressable::URI.normalize_component(
-          self.fragment.strip,
+        component = Addressable::URI.normalize_component(
+          self.fragment,
           Addressable::URI::CharacterClasses::FRAGMENT
         )
+        component == "" ? nil : component
       end)
     end
 
@@ -1771,7 +1805,7 @@ module Addressable
       end
       joined_fragment = uri.fragment
 
-      return Addressable::URI.new(
+      return self.class.new(
         :scheme => joined_scheme,
         :user => joined_user,
         :password => joined_password,
@@ -1827,7 +1861,7 @@ module Addressable
         end
       end
 
-      uri = Addressable::URI.new
+      uri = self.class.new
       uri.defer_validation do
         # Bunch of crazy logic required because of the composite components
         # like userinfo and authority.
@@ -1911,9 +1945,16 @@ module Addressable
               components[:query] = nil
             end
           else
-            if uri.path != SLASH
-              components[:path] = components[:path].gsub(
-                Regexp.new("^" + Regexp.escape(uri.path)), EMPTY_STR)
+            if uri.path != SLASH and components[:path]
+              self_splitted_path = split_path(components[:path])
+              uri_splitted_path = split_path(uri.path)
+              self_dir = self_splitted_path.shift
+              uri_dir = uri_splitted_path.shift
+              while !self_splitted_path.empty? && !uri_splitted_path.empty? and self_dir == uri_dir
+                self_dir = self_splitted_path.shift
+                uri_dir = uri_splitted_path.shift
+              end
+              components[:path] = (uri_splitted_path.fill('..') + [self_dir] + self_splitted_path).join(SLASH)
             end
           end
         end
@@ -1968,7 +2009,7 @@ module Addressable
         end
       end
 
-      return Addressable::URI.new(
+      return self.class.new(
         :scheme => normalized_scheme,
         :authority => normalized_authority,
         :path => normalized_path,
@@ -2065,7 +2106,7 @@ module Addressable
     #
     # @return [Addressable::URI] The cloned URI.
     def dup
-      duplicated_uri = Addressable::URI.new(
+      duplicated_uri = self.class.new(
         :scheme => self.scheme ? self.scheme.dup : nil,
         :user => self.user ? self.user.dup : nil,
         :password => self.password ? self.password.dup : nil,
@@ -2121,6 +2162,15 @@ module Addressable
       replace_self(self.omit(*components))
     end
 
+    ##
+    # Determines if the URI is an empty string.
+    #
+    # @return [TrueClass, FalseClass]
+    #   Returns <code>true</code> if empty, <code>false</code> otherwise.
+    def empty?
+      return self.to_s.empty?
+    end
+
     ##
     # Converts the URI to a <code>String</code>.
     #
@@ -2239,7 +2289,7 @@ module Addressable
     # Ensures that the URI is valid.
     def validate
       return if !!@validation_deferred
-      if self.scheme != nil &&
+      if self.scheme != nil && self.ip_based? &&
           (self.host == nil || self.host.empty?) &&
           (self.path == nil || self.path.empty?)
         raise InvalidURIError,
@@ -2283,5 +2333,19 @@ module Addressable
       @fragment = uri.fragment
       return self
     end
+
+    ##
+    # Splits path string with "/"(slash).
+    # It is considered that there is empty string after last slash when
+    # path ends with slash.
+    #
+    # @param [String] path The path to split.
+    #
+    # @return [Array<String>] An array of parts of path.
+    def split_path(path)
+      splitted = path.split(SLASH)
+      splitted << EMPTY_STR if path.end_with? SLASH
+      splitted
+    end
   end
 end
