@@ -1,3 +1,21 @@
+# coding: utf-8
+# Copyright (C) 2006-2013 Bob Aman
+#
+#    Licensed under the Apache License, Version 2.0 (the "License");
+#    you may not use this file except in compliance with the License.
+#    You may obtain a copy of the License at
+#
+#        http://www.apache.org/licenses/LICENSE-2.0
+#
+#    Unless required by applicable law or agreed to in writing, software
+#    distributed under the License is distributed on an "AS IS" BASIS,
+#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+#    See the License for the specific language governing permissions and
+#    limitations under the License.
+
+
+require "spec_helper"
+
 require "addressable/template"
 
 shared_examples_for 'expands' do |tests|
@@ -14,20 +32,72 @@ shared_examples_for 'expands' do |tests|
   end
 end
 
+describe "eql?" do
+  let(:template) { Addressable::Template.new('https://www.example.com/{foo}') }
+  it 'is equal when the pattern matches' do
+    other_template = Addressable::Template.new('https://www.example.com/{foo}')
+    expect(template).to be_eql(other_template)
+    expect(other_template).to be_eql(template)
+  end
+  it 'is not equal when the pattern differs' do
+    other_template = Addressable::Template.new('https://www.example.com/{bar}')
+    expect(template).to_not be_eql(other_template)
+    expect(other_template).to_not be_eql(template)
+  end
+  it 'is not equal to non-templates' do
+    uri = 'https://www.example.com/foo/bar'
+    addressable_template = Addressable::Template.new uri
+    addressable_uri = Addressable::URI.parse uri
+    expect(addressable_template).to_not be_eql(addressable_uri)
+    expect(addressable_uri).to_not be_eql(addressable_template)
+  end
+end
+
+describe "==" do
+  let(:template) { Addressable::Template.new('https://www.example.com/{foo}') }
+  it 'is equal when the pattern matches' do
+    other_template = Addressable::Template.new('https://www.example.com/{foo}')
+    expect(template).should == other_template
+    expect(other_template).should == template
+  end
+  it 'is not equal when the pattern differs' do
+    other_template = Addressable::Template.new('https://www.example.com/{bar}')
+    expect(template).should_not == other_template
+    expect(other_template).should_not == template
+  end
+  it 'is not equal to non-templates' do
+    uri = 'https://www.example.com/foo/bar'
+    addressable_template = Addressable::Template.new uri
+    addressable_uri = Addressable::URI.parse uri
+    expect(addressable_template).should_not == addressable_uri
+    expect(addressable_uri).should_not == addressable_template
+  end
+end
+
 describe "Type conversion" do
-  subject{
-    {:var => true, :hello => 1234, :nothing => nil, :sym => :symbolic}
+  require "bigdecimal"
+
+  subject {
+    {
+      :var => true,
+      :hello => 1234,
+      :nothing => nil,
+      :sym => :symbolic,
+      :decimal => BigDecimal.new(1)
+    }
   }
+
   it_behaves_like 'expands', {
     '{var}' => 'true',
     '{hello}' => '1234',
     '{nothing}' => '',
-    '{sym}' => 'symbolic'
+    '{sym}' => 'symbolic',
+    '{decimal}' => '0.1E1'
   }
 end
 
 describe "Level 1:" do
-  subject{
+  subject {
     {:var => "value", :hello => "Hello World!"}
   }
   it_behaves_like 'expands', {
@@ -37,7 +107,7 @@ describe "Level 1:" do
 end
 
 describe "Level 2" do
-  subject{
+  subject {
     {
       :var => "value",
       :hello => "Hello World!",
@@ -61,7 +131,7 @@ describe "Level 2" do
 end
 
 describe "Level 3" do
-  subject{
+  subject {
     {
       :var => "value",
       :hello => "Hello World!",
@@ -122,7 +192,7 @@ describe "Level 3" do
 end
 
 describe "Level 4" do
-  subject{
+  subject {
     {
       :var => "value",
       :hello => "Hello World!",
@@ -320,7 +390,7 @@ describe "Level 4" do
   end
 end
 describe "Modifiers" do
-  subject{
+  subject {
     {
       :var => "value",
       :semi => ";",
@@ -345,7 +415,7 @@ describe "Modifiers" do
   end
 end
 describe "Expansion" do
-  subject{
+  subject {
     {
       :count => ["one", "two", "three"],
       :dom => ["example", "com"],
@@ -652,6 +722,11 @@ class ExampleTwoProcessor
   end
 end
 
+class DumbProcessor
+  def self.match(name)
+    return ".*?" if name == "first"
+  end
+end
 
 describe Addressable::Template do
   describe "Matching" do
@@ -669,54 +744,80 @@ describe Addressable::Template do
     let(:uri4){
       Addressable::URI.parse("http://example.com/?a=1&b=2&c=3&first=foo")
     }
+    let(:uri5){
+      "http://example.com/foo"
+    }
     context "first uri with ExampleTwoProcessor" do
-      subject{
+      subject {
         match = Addressable::Template.new(
           "http://example.com/search/{query}/"
         ).match(uri, ExampleTwoProcessor)
       }
-      its(:variables){ should == ["query"]}
-      its(:captures){ should == ["an example search query"]}
+      its(:variables){ should == ["query"] }
+      its(:captures){ should == ["an example search query"] }
     end
 
     context "second uri with ExampleTwoProcessor" do
-      subject{
+      subject {
         match = Addressable::Template.new(
           "http://example.com/{first}/{+second}/"
         ).match(uri2, ExampleTwoProcessor)
       }
-      its(:variables){ should == ["first", "second"]}
+      its(:variables){ should == ["first", "second"] }
       its(:captures){ should == ["a", "b/c"] }
     end
+
+    context "second uri with DumbProcessor" do
+      subject {
+        match = Addressable::Template.new(
+          "http://example.com/{first}/{+second}/"
+        ).match(uri2, DumbProcessor)
+      }
+      its(:variables){ should == ["first", "second"] }
+      its(:captures){ should == ["a", "b/c"] }
+    end
+
     context "second uri" do
-      subject{
+      subject {
         match = Addressable::Template.new(
           "http://example.com/{first}{/second*}/"
         ).match(uri2)
       }
-      its(:variables){ should == ["first", "second"]}
+      its(:variables){ should == ["first", "second"] }
       its(:captures){ should == ["a", ["b","c"]] }
     end
     context "third uri" do
-      subject{
+      subject {
         match = Addressable::Template.new(
           "http://example.com/{;hash*,first}"
         ).match(uri3)
       }
-      its(:variables){ should == ["hash", "first"]}
+      its(:variables){ should == ["hash", "first"] }
       its(:captures){ should == [
         {"a" => "1", "b" => "2", "c" => "3", "first" => "foo"}, nil] }
     end
+    # Note that this expansion is impossible to revert deterministically - the
+    # * operator means first could have been a key of hash or a separate key.
+    # Semantically, a separate key is more likely, but both are possible.
     context "fourth uri" do
-      subject{
+      subject {
         match = Addressable::Template.new(
           "http://example.com/{?hash*,first}"
         ).match(uri4)
       }
-      its(:variables){ should == ["hash", "first"]}
+      its(:variables){ should == ["hash", "first"] }
       its(:captures){ should == [
         {"a" => "1", "b" => "2", "c" => "3", "first"=> "foo"}, nil] }
     end
+    context "fifth uri" do
+      subject {
+        match = Addressable::Template.new(
+          "http://example.com/{path}{?hash*,first}"
+        ).match(uri5)
+      }
+      its(:variables){ should == ["path", "hash", "first"] }
+      its(:captures){ should == ["foo", nil, nil] }
+    end
   end
   describe "extract" do
     let(:template) {
@@ -725,7 +826,8 @@ describe Addressable::Template do
       )
     }
     let(:uri){ "http://example.com/a/b/c/?one=1&two=2#foo" }
-    it "should be able to extract" do
+    let(:uri2){ "http://example.com/a/b/c/#foo" }
+    it "should be able to extract with queries" do
       template.extract(uri).should == {
         "host" => "example.com",
         "segments" => %w(a b c),
@@ -735,10 +837,91 @@ describe Addressable::Template do
         "fragment" => "foo"
       }
     end
+    it "should be able to extract without queries" do
+      template.extract(uri2).should == {
+        "host" => "example.com",
+        "segments" => %w(a b c),
+        "one" => nil,
+        "bogus" => nil,
+        "two" => nil,
+        "fragment" => "foo"
+      }
+    end
+
+    context "issue #137" do
+      subject { Addressable::Template.new('/path{?page,per_page}') }
+
+      it "can match empty" do
+        data = subject.extract("/path")
+        data["page"].should == nil
+        data["per_page"].should == nil
+        data.keys.sort.should == ['page', 'per_page']
+      end
+
+      it "can match first var" do
+        data = subject.extract("/path?page=1")
+        data["page"].should == "1"
+        data["per_page"].should == nil
+        data.keys.sort.should == ['page', 'per_page']
+      end
+
+      it "can match second var" do
+        data = subject.extract("/path?per_page=1")
+        data["page"].should == nil
+        data["per_page"].should == "1"
+        data.keys.sort.should == ['page', 'per_page']
+      end
+
+      it "can match both vars" do
+        data = subject.extract("/path?page=2&per_page=1")
+        data["page"].should == "2"
+        data["per_page"].should == "1"
+        data.keys.sort.should == ['page', 'per_page']
+      end
+    end
   end
-  describe "Partial expand" do
+
+  describe "Partial expand with symbols" do
     context "partial_expand with two simple values" do
-      subject{
+      subject {
+        Addressable::Template.new("http://example.com/{one}/{two}/")
+      }
+      it "builds a new pattern" do
+        subject.partial_expand(:one => "1").pattern.should ==
+          "http://example.com/1/{two}/"
+      end
+    end
+    context "partial_expand query with missing param in middle" do
+      subject {
+        Addressable::Template.new("http://example.com/{?one,two,three}/")
+      }
+      it "builds a new pattern" do
+        subject.partial_expand(:one => "1", :three => "3").pattern.should ==
+          "http://example.com/?one=1{&two}&three=3/"
+      end
+    end
+    context "partial_expand with query string" do
+      subject {
+        Addressable::Template.new("http://example.com/{?two,one}/")
+      }
+      it "builds a new pattern" do
+        subject.partial_expand(:one => "1").pattern.should ==
+          "http://example.com/{?two}&one=1/"
+      end
+    end
+    context "partial_expand with path operator" do
+      subject {
+        Addressable::Template.new("http://example.com{/one,two}/")
+      }
+      it "builds a new pattern" do
+        subject.partial_expand(:one => "1").pattern.should ==
+          "http://example.com/1{/two}/"
+      end
+    end
+  end
+  describe "Partial expand with strings" do
+    context "partial_expand with two simple values" do
+      subject {
         Addressable::Template.new("http://example.com/{one}/{two}/")
       }
       it "builds a new pattern" do
@@ -747,7 +930,7 @@ describe Addressable::Template do
       end
     end
     context "partial_expand query with missing param in middle" do
-      subject{
+      subject {
         Addressable::Template.new("http://example.com/{?one,two,three}/")
       }
       it "builds a new pattern" do
@@ -756,7 +939,7 @@ describe Addressable::Template do
       end
     end
     context "partial_expand with query string" do
-      subject{
+      subject {
         Addressable::Template.new("http://example.com/{?two,one}/")
       }
       it "builds a new pattern" do
@@ -765,7 +948,7 @@ describe Addressable::Template do
       end
     end
     context "partial_expand with path operator" do
-      subject{
+      subject {
         Addressable::Template.new("http://example.com{/one,two}/")
       }
       it "builds a new pattern" do
@@ -776,7 +959,7 @@ describe Addressable::Template do
   end
   describe "Expand" do
     context "expand with a processor" do
-      subject{
+      subject {
         Addressable::Template.new("http://example.com/search/{query}/")
       }
       it "processes spaces" do
@@ -792,7 +975,7 @@ describe Addressable::Template do
       end
     end
     context "partial_expand query with missing param in middle" do
-      subject{
+      subject {
         Addressable::Template.new("http://example.com/{?one,two,three}/")
       }
       it "builds a new pattern" do
@@ -801,7 +984,7 @@ describe Addressable::Template do
       end
     end
     context "partial_expand with query string" do
-      subject{
+      subject {
         Addressable::Template.new("http://example.com/{?two,one}/")
       }
       it "builds a new pattern" do
@@ -810,7 +993,7 @@ describe Addressable::Template do
       end
     end
     context "partial_expand with path operator" do
-      subject{
+      subject {
         Addressable::Template.new("http://example.com{/one,two}/")
       }
       it "builds a new pattern" do
@@ -833,8 +1016,8 @@ describe Addressable::Template do
       end
       it "can match empty" do
         data = subject.match("foo/baz")
-        data.mapping["foo"].should == ""
-        data.mapping["bar"].should == ""
+        data.mapping["foo"].should == nil
+        data.mapping["bar"].should == nil
       end
       it "lists vars" do
         subject.variables.should == ["foo", "bar"]
@@ -848,6 +1031,22 @@ describe Addressable::Template do
         data.mapping["foo"].should == "/test/banana"
         data.mapping["bar"].should == "baz"
       end
+      it "can match empty level 2 #" do
+        data = subject.match("foo/test/bananabaz")
+        data.mapping["foo"].should == "/test/banana"
+        data.mapping["bar"].should == nil
+        data = subject.match("foo/test/banana#baz")
+        data.mapping["foo"].should == "/test/banana"
+        data.mapping["bar"].should == ""
+      end
+      it "can match empty level 2 +" do
+        data = subject.match("foobaz")
+        data.mapping["foo"].should == nil
+        data.mapping["bar"].should == nil
+        data = subject.match("foo#barbaz")
+        data.mapping["foo"].should == nil
+        data.mapping["bar"].should == "bar"
+      end
       it "lists vars" do
         subject.variables.should == ["foo", "bar"]
       end
@@ -922,6 +1121,39 @@ describe Addressable::Template do
             subject.variables.should == %w(foo bar)
           end
         end
+
+        context "issue #137" do
+          subject { Addressable::Template.new('/path{?page,per_page}') }
+
+          it "can match empty" do
+            data = subject.match("/path")
+            data.mapping["page"].should == nil
+            data.mapping["per_page"].should == nil
+            data.mapping.keys.sort.should == ['page', 'per_page']
+          end
+
+          it "can match first var" do
+            data = subject.match("/path?page=1")
+            data.mapping["page"].should == "1"
+            data.mapping["per_page"].should == nil
+            data.mapping.keys.sort.should == ['page', 'per_page']
+          end
+
+          it "can match second var" do
+            data = subject.match("/path?per_page=1")
+            data.mapping["page"].should == nil
+            data.mapping["per_page"].should == "1"
+            data.mapping.keys.sort.should == ['page', 'per_page']
+          end
+
+          it "can match both vars" do
+            data = subject.match("/path?page=2&per_page=1")
+            data.mapping["page"].should == "2"
+            data.mapping["per_page"].should == "1"
+            data.mapping.keys.sort.should == ['page', 'per_page']
+          end
+        end
+
         context "issue #71" do
           subject { Addressable::Template.new("http://cyberscore.dev/api/users{?username}") }
           it "can match" do
@@ -1028,10 +1260,10 @@ describe Addressable::Template::MatchData do
   its(:template) { should == template }
   its(:mapping) { should == { 'foo' => 'ab', 'bar' => 'cd' } }
   its(:variables) { should == ['foo', 'bar'] }
-  its(:keys) { should == its.variables }
-  its(:names) { should == its.variables }
+  its(:keys) { should == ['foo', 'bar'] }
+  its(:names) { should == ['foo', 'bar'] }
   its(:values) { should == ['ab', 'cd'] }
-  its(:captures) { should == its.values }
+  its(:captures) { should == ['ab', 'cd'] }
   its(:to_a) { should == ['ab/cd', 'ab', 'cd'] }
   its(:to_s) { should == 'ab/cd' }
   its(:string) { should == its.to_s }
